from __future__ import annotations

import sys
import typing as t
from types import ModuleType

import pytest

import markupsafe
from markupsafe import _native

try:
    from markupsafe import _speedups
except ImportError:
    _speedups = None  # type: ignore


def pytest_report_header() -> list[str]:
    """Return a list of strings to be displayed in the header of the report."""
    if sys.version_info >= (3, 13):
        return [f"Free-threaded: {not sys._is_gil_enabled()}"]

    return []


@pytest.fixture(
    scope="session",
    autouse=True,
    params=(
        _native,
        pytest.param(
            _speedups,
            marks=pytest.mark.skipif(_speedups is None, reason="speedups unavailable"),
        ),
    ),
)
def _mod(request: pytest.FixtureRequest) -> None:
    mod = t.cast(ModuleType, request.param)
    markupsafe._escape_inner = mod._escape_inner  # type: ignore[attr-defined]



import subprocess
import sys

for name, s in (
    ("short escape", '"<strong>Hello, World!</strong>"'),
    ("long escape", '"Hello, World!" * 1000'),
    ("short plain", '"Hello, World!"'),
    ("long plain", '"Hello, World!" * 1000'),
    ("long suffix", '"<strong>Hello, World!</strong>" + "x" * 100_000'),
):
    for mod in "native", "speedups":
        subprocess.run(
            [
                sys.executable,
                "-m",
                "pyperf",
                "timeit",
                "--name",
                f"{name} {mod}",
                "-s",
                (
                    "import markupsafe\n"
                    f"from markupsafe._{mod} import _escape_inner\n"
                    "markupsafe._escape_inner = _escape_inner\n"
                    "from markupsafe import escape\n"
                    f"s = {s}"
                ),
                "escape(s)",
            ]
        )



from __future__ import annotations

import pytest

from markupsafe import escape


class CustomHtmlThatRaises:
    def __html__(self) -> str:
        raise ValueError(123)


def test_exception_custom_html() -> None:
    """Checks whether exceptions in custom __html__ implementations are
    propagated correctly.

    There was a bug in the native implementation at some point:
    https://github.com/pallets/markupsafe/issues/108
    """
    obj = CustomHtmlThatRaises()

    with pytest.raises(ValueError):
        escape(obj)



from __future__ import annotations

import typing as t

import pytest

from markupsafe import escape
from markupsafe import Markup


@pytest.mark.parametrize(
    ("value", "expect"),
    (
        # empty
        ("", ""),
        # ascii
        ("abcd&><'\"efgh", "abcd&amp;&gt;&lt;&#39;&#34;efgh"),
        ("&><'\"efgh", "&amp;&gt;&lt;&#39;&#34;efgh"),
        ("abcd&><'\"", "abcd&amp;&gt;&lt;&#39;&#34;"),
        # 2 byte
        ("こんにちは&><'\"こんばんは", "こんにちは&amp;&gt;&lt;&#39;&#34;こんばんは"),
        ("&><'\"こんばんは", "&amp;&gt;&lt;&#39;&#34;こんばんは"),
        ("こんにちは&><'\"", "こんにちは&amp;&gt;&lt;&#39;&#34;"),
        # 4 byte
        (
            "\U0001f363\U0001f362&><'\"\U0001f37a xyz",
            "\U0001f363\U0001f362&amp;&gt;&lt;&#39;&#34;\U0001f37a xyz",
        ),
        ("&><'\"\U0001f37a xyz", "&amp;&gt;&lt;&#39;&#34;\U0001f37a xyz"),
        ("\U0001f363\U0001f362&><'\"", "\U0001f363\U0001f362&amp;&gt;&lt;&#39;&#34;"),
    ),
)
def test_escape(value: str, expect: str) -> None:
    assert escape(value) == Markup(expect)


class Proxy:
    def __init__(self, value: t.Any) -> None:
        self.__value = value

    @property  # type: ignore[misc]
    def __class__(self) -> type[t.Any]:
        # Make o.__class__ and isinstance(o, str) see the proxied object.
        return self.__value.__class__  # type: ignore[no-any-return]

    def __str__(self) -> str:
        return str(self.__value)


def test_proxy() -> None:
    """Handle a proxy object that pretends its __class__ is str."""
    p = Proxy("test")
    assert p.__class__ is str
    assert isinstance(p, str)
    assert escape(p) == Markup("test")


class ReferenceStr(str):
    def __str__(self) -> str:
        # This should return a str, but it returns the subclass instead.
        return self


def test_subclass() -> None:
    """Handle if str(o) does not return a plain str."""
    s = ReferenceStr("test")
    assert isinstance(s, str)
    assert escape(s) == Markup("test")



from __future__ import annotations

import gc

from markupsafe import escape


def test_markup_leaks() -> None:
    counts = set()
    # Try to start with a "clean" count. Works for PyPy but not 3.13 JIT.
    gc.collect()

    for _ in range(20):
        for _ in range(1000):
            escape("foo")
            escape("<foo>")
            escape("foo")
            escape("<foo>")

        counts.add(len(gc.get_objects()))

    # Some implementations, such as PyPy and Python 3.13 JIT, end up with 2
    # counts rather than one. Presumably this is internals stabilizing. A leak
    # would presumably have a different count every loop.
    assert len(counts) < 3
